import nltk

# nltk.download()

"""

Analyze text by calculating different scores for sentences.
"""

# number of words
min_sentence_length = 2
max_sentence_length = 24

# number of symbols
min_word_length = 1
max_word_length = 9

# percent of complex vocabulary
min_complex_vocabulary = 0.04
max_complex_vocabulary = 0.82

# path to dictionary with 1000 most frequent english terms
dictionary = "../res/dictionary.txt"
complex_words = ""

# ratio of nouns and nominalized words to verbs
min_nominal_ratio = 0.21
max_nominal_ratio = 0.94

# number of sub-sentences or side information
min_sub_sentences = 1
max_sub_sentences = 5

# percent of imprecise forms
min_imprecise_forms = 0.04
max_imprecise_forms = 0.42

# path to dictionary with imprecise english terms
imprecise = "../res/imprecise.txt"
imprecise_words = ""

# score limits for annotation text
limit_high = 0.85
limit_very_high = 0.95


def text_analysis(sentences):
    """
    Analyze multiple sentences, set an annotation text for feature scores that hit a predefined limit and return a
    matrix with scores for all attributes.

    :param sentences:   all sentences to analyze
    :return:    a score matrix *n x m*, where *n* is the number of sentences and *m* is the number of attributes
    """

    scn = 6  # number of scores to calculate

    w, h = scn, len(sentences)
    matrix_scores = [[0 for x in range(w)] for y in range(h)]

    for i in range(len(sentences)):
        annotation_text = ""
        prefix = "• The sentence '" + ' '.join(sentences[i].sentence.split()[:3]) + " [...]' "
        prefix_short = "• It "

        matrix_scores[i][0] = calculate_vocabulary_complexity_score(sentences[i].sentence)

        if matrix_scores[i][0] >= limit_very_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "contains too many complex words." \
                                                " Consider rewriting it.\n"
            annotation_text = annotation_text + "Complex words detected: " + complex_words + "\n"
        elif matrix_scores[i][0] >= limit_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "contains multiple complex words." \
                                                " Consider rewriting it.\n"
            annotation_text = annotation_text + "Complex words detected: " + complex_words + "\n"

        matrix_scores[i][1] = calculate_word_length_score(sentences[i].sentence)

        if matrix_scores[i][1] >= limit_very_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "contains too many long words." \
                                                " Consider replacing them.\n"
        elif matrix_scores[i][1] >= limit_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "contains multiple long words." \
                                                " Consider replacing them.\n"

        matrix_scores[i][2] = calculate_nominal_form_score(sentences[i].sentence)

        if matrix_scores[i][2] >= limit_very_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "contains too many nouns or gerunds." \
                                                " Consider reformulating it.\n"
        elif matrix_scores[i][2] >= limit_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "contains multiple nouns or gerunds." \
                                                " Consider reformulating it.\n"

        matrix_scores[i][3] = calculate_sentence_length_score(sentences[i].sentence)

        if matrix_scores[i][3] >= limit_very_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "is too long. Consider shortening it or split it up in" \
                                                " two or more sentences.\n"
        elif matrix_scores[i][3] >= limit_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "is long. Consider shortening it.\n"

        matrix_scores[i][4] = calculate_sentence_structure_score(sentences[i].sentence)

        if matrix_scores[i][4] >= limit_very_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "has a too complex structure. Consider reformulating" \
                                                " it or remove nested sub-sentences.\n"
        elif matrix_scores[i][4] >= limit_high:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "has a complex structure. Consider reformulating it" \
                                                " or remove nested sub-sentences.\n"

        matrix_scores[i][5] = calculate_imprecise_form_score(sentences[i].sentence)
        if matrix_scores[i][5] >= min_imprecise_forms:
            annotation_text = annotation_text + (prefix if len(annotation_text) < 1 else prefix_short) +\
                                                "contains imprecise formulations. Consider replacing them with" \
                                                " precise terms.\n"
            annotation_text = annotation_text + "Imprecise formulations detected: " + imprecise_words + "\n"

        if len(annotation_text) > 0:
            sentences[i].annotText = annotation_text

    return matrix_scores


def analyze_single_sentence(sentence):
    """
    Analyze a single sentence and return a one dimensional matrix with scores for all attributes of this sentence.

    :param sentence:     the single sentence to analyze
    :return:     a score matrix *1 x m*, where *m* is the number of attributes
    """
    scn = 6  # number of scores to calculate
    w, h = scn, 1
    matrix_scores = [[0 for x in range(w)] for y in range(h)]
    matrix_scores[0][0] = calculate_vocabulary_complexity_score(sentence)
    matrix_scores[0][1] = calculate_word_length_score(sentence)
    matrix_scores[0][2] = calculate_nominal_form_score(sentence)
    matrix_scores[0][3] = calculate_sentence_length_score(sentence)
    matrix_scores[0][4] = calculate_sentence_structure_score(sentence)
    matrix_scores[0][5] = calculate_imprecise_form_score(sentence)

    return matrix_scores


def calculate_vocabulary_complexity_score(sentence):
    """
    Calculate the Vocabulary Complexity (VC). This is the percentage of terms in a sentence, not contained in a list of
    the *1000* most frequent terms in english language.

    :param sentence:    the single sentence to analyze
    :return:    a value that represents the VC score of the sentence
    """

    if len(sentence.split()) <= 0:
        # not all sentences contain text
        return 0

    global complex_words
    complex_words = ""
    complex_voc = 0
    punctuation = ".,:;"
    file = open(dictionary, 'r+')
    contents = file.read()
    for i in range(len(sentence.split())):
        search_word = replace_punctuation(sentence.split()[i])
        if search_word not in punctuation and search_word.lower() not in contents.lower()\
                and not search_word.startswith("[") and not search_word.isdigit():
            # print(search_word)
            complex_words = complex_words + ("" if len(complex_words) < 1 else ", ") + search_word
            complex_voc = complex_voc + 1
    file.close()
    # print(sentence)
    # print(complex_voc)
    percentage = complex_voc / len(sentence.split())
    # print(percentage)
    vocabulary_complexity_score = map_to_score(percentage, min_complex_vocabulary, max_complex_vocabulary)
    # print(vocabulary_complexity_score)

    return vocabulary_complexity_score


def calculate_word_length_score(sentence):
    """
    Calculate the Word Length (WL). This is the average number of characters in a word.

    :param sentence:    the single sentence to analyze
    :return:    a value that represents the WL score of the sentence
    """
    if len(sentence.split()) <= 0:
        # not all sentences contain text
        return 0

    # number of symbols without blanks and punctuation / number of words
    average_word_length = len(replace_punctuation(sentence)) / len(sentence.split())
    # print("sentence: ", sentence)
    # print("average word length: ", average_word_length)
    word_length_score = map_to_score(average_word_length, min_word_length, max_word_length)
    # print(word_length_score)

    return word_length_score


def calculate_nominal_form_score(sentence):
    """
    Calculate the Nominal Forms (NF). This is the combination of the noun-to-verb ratio and the number of nominal forms.
    Nominal forms include gerunds, nominalized words and nouns. Nominalized words contain words with the endings *ing*,
    *ity*, *ness*, and similar.

    :param sentence:    the single sentence to analyze
    :return:    a value that represents the NF score of the sentence
    """
    if len(sentence.strip()) <= 0:
        return 0

    # number of nouns in sentence
    token_text = nltk.word_tokenize(sentence)
    tagged_sent = nltk.pos_tag(token_text)
    nouns = []
    verbs = []
    nominalized = []
    for word, pos in tagged_sent:
        if pos.startswith('N'):
            # NN noun, singular 'desk'
            # NNS noun plural 'desks'
            # NNP proper noun, singular 'Harrison'
            # NNPS proper noun, plural 'Americans'
            nouns.append(word)
        elif pos.startswith('VBG'):
            # VBG verb, gerund/present participle 'taking'
            nominalized.append(word)
        elif pos.startswith('V'):
            # VB verb, base form 'take'
            # VBD verb, past tense 'took'
            # VBN verb, past participle 'taken'
            # VBP verb, sing. present, non-3d 'take'
            # VBZ verb, 3rd person sing. present 'takes'
            verbs.append(word)
    noun_count = len(nouns)
    verb_count = len(verbs)
    nominalized_count = len(nominalized)

    if noun_count + verb_count + nominalized_count <= 0:
        return 0

    ratio = (noun_count + nominalized_count) / (noun_count + verb_count + nominalized_count)
    # print("ratio: ", ratio)

    nominal_form_score = map_to_score(ratio, min_nominal_ratio, max_nominal_ratio)
    # print(nominal_form_score)

    return nominal_form_score


def calculate_sentence_length_score(sentence):
    """
    Calculate the Sentence Length (SL). This is defined by the number of words in a sentence.

    :param sentence:    the single sentence to analyze
    :return:    a value that represents the SL score of the sentence
    """
    # number of words in a sentence
    # print("sentence: ", sentence)
    sentence_length = len(sentence.split())
    # print("sentence length: ", sentence_length)
    sentence_length_score = map_to_score(sentence_length, min_sentence_length, max_sentence_length)
    # print(sentence_length_score)

    return sentence_length_score


def calculate_sentence_structure_score(sentence):
    """
    Calculate the Sentence Structure (SS). The complexity is measured by branching in the sentence tree. It is increased
    when the sentence is interrupted by sub-sentences or parenthesis.

    :param sentence:    the single sentence to analyze
    :return:    a value that represents the SS score of the sentence
    """
    tokenizer = nltk.data.load('tokenizers/punkt/english.pickle')
    data = sentence.replace(',', '.')
    data = data.replace(';', '.')
    data = data.replace('(', '.')
    data = data.replace('for', '.')
    data = data.replace('and', '.')
    data = data.replace('nor', '.')
    data = data.replace('but', '.')
    data = data.replace('or', '.')
    data = data.replace('yet', '.')
    data = data.replace('so', '.')
    data = data.replace('though', '.')
    data = data.replace('while', '.')
    data = data.replace('if', '.')
    # print(sentence)

    # number of sub-sentences or side information in brackets
    sub_sentence_count = len(tokenizer.tokenize(data))
    # print("count: ", sub_sentence_count)
    sentence_structure_score = map_to_score(sub_sentence_count, min_sub_sentences, max_sub_sentences)
    # print(sentence_structure_score)

    return sentence_structure_score


def calculate_imprecise_form_score(sentence):
    """
    Calculate the Imprecise Form (IF). It is measured by the percentage of terms contained in a list of imprecise words
    that are used in a sentence.

    :param sentence:    the single sentence to analyze
    :return:    a value that represents the IF score of the sentence
    """

    if len(sentence.split()) <= 0:
        # not all sentences contain text
        return 0

    global imprecise_words
    imprecise_words = ""
    imprecise_voc = 0
    file = open(imprecise, 'r+')
    contents = file.read()
    for i in range(len(sentence.split())):
        search_word = replace_punctuation(sentence.split()[i])
        for line in contents.splitlines():
            if (line.lower() == search_word.lower()) and (line.lower() not in imprecise_words.lower()):
                # print(search_word)
                imprecise_words = imprecise_words + ("" if len(imprecise_words) < 1 else ", ") + search_word
                imprecise_voc = imprecise_voc + 1
            if (" " + line.lower() in sentence.lower()) and (line.lower() not in imprecise_words.lower()):
                imprecise_words = imprecise_words + ("" if len(imprecise_words) < 1 else ", ") + line
                imprecise_voc = imprecise_voc + 1
    file.close()
    # print("sentence: ", sentence)
    # print(imprecise_voc)
    # print("imprecise: ", imprecise_words)
    percentage = imprecise_voc / len(sentence.split())
    # print(percentage)
    imprecise_form_score = map_to_score(percentage, min_imprecise_forms, max_imprecise_forms)
    # print(imprecise_form_score)

    return imprecise_form_score


def map_to_score(value, min_limit, max_limit):
    """
    Map a value from *[min_limit  max_limit]* to the interval *[0  1]* and return a score value.

    :param value:   the value to map
    :param min_limit:   the min limit of the value
    :param max_limit:   the max limit of the value
    :return:    the mapped score value in the interval *[0  1]*
    """
    if value < min_limit:
        return 0
    if value > max_limit:
        return 1
    return (value - min_limit) / (max_limit - min_limit)


def replace_punctuation(sentence):
    """
    Replace the punctuation in a sentence

    :param sentence:    the sentence to process
    :return:    the processed sentence without spaces and without punctuation
    """
    sen = sentence
    sen = sen.replace(" ", "")
    sen = sen.replace(".", "")
    sen = sen.replace("!", "")
    sen = sen.replace("?", "")
    sen = sen.replace(":", "")
    sen = sen.replace(",", "")
    sen = sen.replace(";", "")
    sen = sen.replace("(", "")
    sen = sen.replace(")", "")
    return sen
