# -*- coding: utf-8 -*-

import ntpath
import os
import unicodedata
import webbrowser

import math
import nltk
import datetime
import fitz
import pdfminer
from threading import Thread
from TextAnalyzer.src.analyzeText import text_analysis, analyze_single_sentence
from TextAnalyzer.src.pdfHandler import PdfHandler
from TextAnalyzer.src.overlay import Overlay
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import QPoint, Qt, QTimer
from PyQt5.QtGui import QPixmap, QPainter, QColor, QPalette, QBrush, QPen
from PyQt5.QtWidgets import QTableWidgetItem, QLabel, QWidget, QTextEdit, QGridLayout, QPushButton, QDesktopWidget

initialized = 0
scores = 0
ui_e = 0
sent = 0

cell_row = 0
cell_column = 0
cell_text = 0

# define colors
color_hard = (0.99, 0.36, 0.36)
color_easy = (0.46, 0.46, 0.99)
color_neutral = (1.0, 1.0, 1.0)

css = "../ui/style.css"
pdf_path = ""


def vocabulary_slider_value_change(value):
    """
    Called when the user changes the vocabulary slider in the ui. This leads to an update of the sentence score values
    and of the sentence color values.

    :param value:   the updated value of the vocabulary slider
    :return:    None
    """
    print('Vocabulary slider: ', (value * 10 - 10) / 100)
    if initialized:
        update_colors()


def word_length_slider_value_change(value):
    """
    Called when the user changes the word length slider in the ui. This leads to an update of the sentence score values
    and of the sentence color values.

    :param value:   the updated value of the word length slider
    :return:    None
    """
    print('Word length slider: ', (value * 10 - 10) / 100)
    if initialized:
        update_colors()


def nominal_form_slider_value_change(value):
    """
    Called when the user changes the nominal form slider in the ui. This leads to an update of the sentence score values
    and of the sentence color values.

    :param value:   the updated value of the nominal form slider
    :return:    None
    """
    print('Nominal form slider: ', (value * 10 - 10) / 100)
    if initialized:
        update_colors()


def sentence_length_slider_value_change(value):
    """
    Called when the user changes the sentence length slider in the ui. This leads to an update of the sentence score
    values and of the sentence color values.

    :param value:   the updated value of the sentence length slider
    :return:    None
    """
    print('Sentence length slider: ', (value * 10 - 10) / 100)
    if initialized:
        update_colors()


def sentence_structure_slider_value_change(value):
    """
    Called when the user changes the sentence structure slider in the ui. This leads to an update of the sentence score
    values and of the sentence color values.

    :param value:   the updated value of the sentence structure slider
    :return:    None
    """
    print('Sentence structure slider: ', (value * 10 - 10) / 100)
    if initialized:
        update_colors()


def imprecise_form_slider_value_change(value):
    """
    Called when the user changes the imprecise form slider in the ui. This leads to an update of the sentence score
    values and of the sentence color values.

    :param value:   the updated value of the imprecise form slider
    :return:    None
    """
    print('Imprecise form slider: ', (value * 10 - 10) / 100)
    if initialized:
        update_colors()


def interpolate_color(color2, color1, t):
    """
    Interpolate between two colors.

    :param color2:  one color for the interpolation
    :param color1:  another color for the interpolation
    :param t:   the interpolation coefficient
    :return:    the interpolated color value
    """
    if color2 < color1:
        ts = t * 2
        x = color1[0] * ts + color2[0] * (1 - ts)
        y = color1[1] * ts + color2[1] * (1 - ts)
        z = color1[2] * ts + color2[2] * (1 - ts)
    else:
        ts = (t - 0.5) * 2
        x = color1[0] * ts + color2[0] * (1 - ts)
        y = color1[1] * ts + color2[1] * (1 - ts)
        z = color1[2] * ts + color2[2] * (1 - ts)

    # print("x, y, z : ", x, ", ", y, ", ", z)
    return [x, y, z]


def slider_count():
    """
    Count how many sliders have an effect on the result. These are all sliders with a weight higher than zero.

    :return:    the number of sliders with a weight higher than zero
    """
    count = 0
    global ui_e
    if ui_e.vocabularySlider.value() > 1:
        count = count + 1
    if ui_e.wordLengthSlider.value() > 1:
        count = count + 1
    if ui_e.nominalFormSlider.value() > 1:
        count = count + 1
    if ui_e.sentenceLengthSlider.value() > 1:
        count = count + 1
    if ui_e.sentenceStructureSlider.value() > 1:
        count = count + 1
    if ui_e.impreciseFormSlider.value() > 1:
        count = count + 1
    return count


def calculate_sentence_score(score_matrix, row):
    """
    Calculate the score of a single sentence.

    :param score_matrix:    the matrix with the score values
    :param row: the row of the sentence, whose score is calculated
    :return:    the calculated score of the sentence
    """
    global ui_e
    if slider_count() == 0:
        sentence_score = 0.5
    else:
        sentence_score = ((score_matrix[row][0] * (ui_e.vocabularySlider.value() * 10 - 10) / 100
                           + score_matrix[row][1] * (ui_e.wordLengthSlider.value() * 10 - 10) / 100
                           + score_matrix[row][2] * (ui_e.nominalFormSlider.value() * 10 - 10) / 100
                           + score_matrix[row][3] * (ui_e.sentenceLengthSlider.value() * 10 - 10) / 100
                           + score_matrix[row][4] * (ui_e.sentenceStructureSlider.value() * 10 - 10) / 100
                           + score_matrix[row][5] * (ui_e.impreciseFormSlider.value() * 10 - 10) / 100)
                          / slider_count())
    # print(sentence_score)
    return sentence_score


def interpolate_sentence_color(sentence_score, sentence):
    """
    Interpolate between colors based on the sentence score.

    :param sentence_score:  the score of the sentence
    :param sentence:    the sentence whose color is interpolated
    :return:    None
    """
    if sentence_score > 0.5:
        sentence.setColor(interpolate_color(color_neutral, color_hard, sentence_score))
    if sentence_score < 0.5:
        sentence.setColor(interpolate_color(color_easy, color_neutral, sentence_score))
    if sentence_score == 0.5:
        sentence.setColor(color_neutral)


def interpolate_score_color(score, sentence):
    """
    Interpolate between colors based on the feature score.

    :param score:  the score of the feature
    :param sentence:    the sentence whose feature color is interpolated
    :return:    None
    """
    interpolate_sentence_color(score, sentence)


def update_sentence_color(row, sentence):
    """
    Update the color of a sentence.

    :param row: the row of the sentence to update
    :param sentence:    the sentence whose color is updated
    :return:    None
    """
    update_score_color(row, 0, sentence)


def update_score_color(row, column, sentence):
    """
    Update the color of a feature.

    :param row: the row of the feature to update
    :param column:  the column of the feature to update
    :param sentence:    the sentence whose feature color is updated
    :return:    None
    """
    global ui_e
    ui_e.detailedView.item(row, column).setBackground(
        QtGui.QColor(sentence.color[0] * 255, sentence.color[1] * 255, sentence.color[2] * 255))


def update_overall_score(sum_score, num):
    """
    Calculate an overall score for the document and show it in the ui.

    :param sum_score:   the sum of all sentence score values
    :param num:   the number of sentences
    :return:    None
    """
    global ui_e
    overall_score = sum_score / num if num > 0 else 0.5
    print("Overall score: ", overall_score)

    ui_e.overall_score_value.setText(str("%.2f" % round(overall_score, 2)))
    ui_e.overall_score_feedback.setText("Neutral")
    if overall_score > 0.80:
        ui_e.overall_score_feedback.setText("Very Hard")
    elif overall_score > 0.65:
        ui_e.overall_score_feedback.setText("Hard")
    elif overall_score < 0.20:
        ui_e.overall_score_feedback.setText("Very Easy")
    elif overall_score < 0.35:
        ui_e.overall_score_feedback.setText("Easy")

    color = color_neutral
    if overall_score > 0.5:
        color = interpolate_color(color_neutral, color_hard, overall_score)
    if overall_score < 0.5:
        color = interpolate_color(color_easy, color_neutral, overall_score)

    ui_e.overall_score_feedback.setAutoFillBackground(True)
    ui_e.overall_score_label.setAutoFillBackground(True)
    ui_e.overall_score_value.setAutoFillBackground(True)
    color = QtGui.QColor(color[0] * 255, color[1] * 255, color[2] * 255)
    alpha = 1
    values = "{r}, {g}, {b}, {a}".format(r=color.red(), g=color.green(), b=color.blue(), a=alpha)
    ui_e.overall_score_feedback.setStyleSheet("QLabel { background-color: rgba(" + values + "); }")
    ui_e.overall_score_label.setStyleSheet("QLabel { background-color: rgba(" + values + "); }")
    ui_e.overall_score_value.setStyleSheet("QLabel { background-color: rgba(" + values + "); }")

    ui_e.overall_score_value.setVisible(True)
    ui_e.overall_score_label.setVisible(True)
    ui_e.overall_score_feedback.setVisible(True)


def update_colors():
    """
    Update the colors of all sentences. First calculate an updated score value. Then interpolate the new sentence
    colors based on the scores calculated. Finally update the sentence colors.

    :return:    None
    """
    global scores, ui_e, sent, initialized
    initialized = 0

    sum_score = 0
    for k, s in enumerate(sent):
        sentence_score = calculate_sentence_score(scores, k)
        interpolate_sentence_color(sentence_score, s)
        update_sentence_color(k, s)
        sum_score = sum_score + sentence_score
    initialized = 1
    ui_e.generatePDFButton.setEnabled(True)
    update_overall_score(sum_score, len(sent))


def set_style_sheet(widget):
    """
    Set style sheet for a widget.

    :param widget: the widget to be styled
    :return:    None
    """
    with open(css, "r") as fh:
        widget.setStyleSheet(fh.read())
    fh.close()


class Ui_MainWindow(object):
    """
    The main window of the user interface. It defines all ui elements.

    """

    def setupUi(self, MainWindow):
        """
        Setup all ui elements.

        :param MainWindow:  the main window of the ui
        :return:    None
        """
        # current handling pdf
        self.handler = None

        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1412, 813)
        # uncomment, if window should be maximized on startup
        # MainWindow.showMaximized()

        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.rightView = QtWidgets.QVBoxLayout()
        self.rightView.setObjectName("rightView")

        self.img_table = QLabel(self.centralwidget)
        self.img_table.setObjectName("img_table")
        self.rightView.addWidget(self.img_table)

        self.detailedView = QtWidgets.QTableWidget(self.centralwidget)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        self.detailedView.setFont(font)
        self.detailedView.setObjectName("detailedView")
        self.detailedView.setColumnCount(7)
        self.detailedView.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.detailedView.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.detailedView.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.detailedView.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.detailedView.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.detailedView.setHorizontalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.detailedView.setHorizontalHeaderItem(5, item)
        item = QtWidgets.QTableWidgetItem()
        self.detailedView.setHorizontalHeaderItem(6, item)
        self.rightView.addWidget(self.detailedView)
        self.horizontalLayout.addLayout(self.rightView)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1434, 21))
        self.menubar.setObjectName("menubar")
        self.menuMenu = QtWidgets.QMenu(self.menubar)
        self.menuMenu.setObjectName("menuMenu")
        MainWindow.setMenuBar(self.menubar)

        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.dockWidget = QtWidgets.QDockWidget(MainWindow)
        self.dockWidget.setObjectName("dockWidget")
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.formLayout = QtWidgets.QFormLayout(self.dockWidgetContents)
        self.formLayout.setObjectName("formLayout")

        global ui_e
        ui_e = self

        self.vocabularySlider = QtWidgets.QSlider(self.dockWidgetContents)
        self.vocabularySlider.setMaximum(10)
        self.vocabularySlider.setSliderPosition(6)
        self.vocabularySlider.setMinimum(1)
        self.vocabularySlider.setMaximum(11)
        self.vocabularySlider.setTickInterval(1)
        self.vocabularySlider.setSingleStep(1)  # arrow-key step-size
        self.vocabularySlider.setPageStep(1)  # mouse-wheel/page-key step-size
        self.vocabularySlider.valueChanged.connect(vocabulary_slider_value_change)

        self.vocabularySlider.setOrientation(QtCore.Qt.Horizontal)
        self.vocabularySlider.setObjectName("vocabularySlider")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.vocabularySlider)

        self.wordLengthSlider = QtWidgets.QSlider(self.dockWidgetContents)
        self.wordLengthSlider.setMaximum(10)
        self.wordLengthSlider.setSliderPosition(9)
        self.wordLengthSlider.setMinimum(1)
        self.wordLengthSlider.setMaximum(11)
        self.wordLengthSlider.setTickInterval(1)
        self.wordLengthSlider.setSingleStep(1)  # arrow-key step-size
        self.wordLengthSlider.setPageStep(1)  # mouse-wheel/page-key step-size
        self.wordLengthSlider.valueChanged.connect(word_length_slider_value_change)

        self.wordLengthSlider.setOrientation(QtCore.Qt.Horizontal)
        self.wordLengthSlider.setObjectName("wordLengthSlider")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.wordLengthSlider)

        self.nominalFormSlider = QtWidgets.QSlider(self.dockWidgetContents)
        self.nominalFormSlider.setMaximum(10)
        self.nominalFormSlider.setSliderPosition(7)
        self.nominalFormSlider.setMinimum(1)
        self.nominalFormSlider.setMaximum(11)
        self.nominalFormSlider.setTickInterval(1)
        self.nominalFormSlider.setSingleStep(1)  # arrow-key step-size
        self.nominalFormSlider.setPageStep(1)  # mouse-wheel/page-key step-size
        self.nominalFormSlider.valueChanged.connect(nominal_form_slider_value_change)

        self.nominalFormSlider.setOrientation(QtCore.Qt.Horizontal)
        self.nominalFormSlider.setObjectName("nominalFormSlider")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.nominalFormSlider)

        self.sentenceLengthSlider = QtWidgets.QSlider(self.dockWidgetContents)
        self.sentenceLengthSlider.setMaximum(10)
        self.sentenceLengthSlider.setSliderPosition(10)
        self.sentenceLengthSlider.setMinimum(1)
        self.sentenceLengthSlider.setMaximum(11)
        self.sentenceLengthSlider.setTickInterval(1)
        self.sentenceLengthSlider.setSingleStep(1)  # arrow-key step-size
        self.sentenceLengthSlider.setPageStep(1)  # mouse-wheel/page-key step-size
        self.sentenceLengthSlider.valueChanged.connect(sentence_length_slider_value_change)

        self.sentenceLengthSlider.setOrientation(QtCore.Qt.Horizontal)
        self.sentenceLengthSlider.setObjectName("sentenceLengthSlider")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.sentenceLengthSlider)

        self.sentenceStructureSlider = QtWidgets.QSlider(self.dockWidgetContents)
        self.sentenceStructureSlider.setMaximum(10)
        self.sentenceStructureSlider.setSliderPosition(9)
        self.sentenceStructureSlider.setMinimum(1)
        self.sentenceStructureSlider.setMaximum(11)
        self.sentenceStructureSlider.setTickInterval(1)
        self.sentenceStructureSlider.setSingleStep(1)  # arrow-key step-size
        self.sentenceStructureSlider.setPageStep(1)  # mouse-wheel/page-key step-size
        self.sentenceStructureSlider.valueChanged.connect(sentence_structure_slider_value_change)

        self.sentenceStructureSlider.setOrientation(QtCore.Qt.Horizontal)
        self.sentenceStructureSlider.setObjectName("sentenceStructureSlider")
        self.formLayout.setWidget(8, QtWidgets.QFormLayout.FieldRole, self.sentenceStructureSlider)

        self.impreciseFormSlider = QtWidgets.QSlider(self.dockWidgetContents)
        self.impreciseFormSlider.setMaximum(10)
        self.impreciseFormSlider.setSliderPosition(7)
        self.impreciseFormSlider.setMinimum(1)
        self.impreciseFormSlider.setMaximum(11)
        self.impreciseFormSlider.setTickInterval(1)
        self.impreciseFormSlider.setSingleStep(1)  # arrow-key step-size
        self.impreciseFormSlider.setPageStep(1)  # mouse-wheel/page-key step-size
        self.impreciseFormSlider.valueChanged.connect(imprecise_form_slider_value_change)

        self.impreciseFormSlider.setOrientation(QtCore.Qt.Horizontal)
        self.impreciseFormSlider.setObjectName("impreciseFormSlider")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.FieldRole, self.impreciseFormSlider)

        self.vocabularyLabel = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.vocabularyLabel.setFont(font)
        self.vocabularyLabel.setObjectName("vocabularyLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.vocabularyLabel)

        self.wordLengthLabel = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.wordLengthLabel.setFont(font)
        self.wordLengthLabel.setObjectName("wordLengthLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.wordLengthLabel)

        self.nominalFormLabel = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.nominalFormLabel.setFont(font)
        self.nominalFormLabel.setObjectName("nominalFormLabel")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.nominalFormLabel)

        self.sentenceLengthLabel = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.sentenceLengthLabel.setFont(font)
        self.sentenceLengthLabel.setObjectName("sentenceLengthLabel")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.sentenceLengthLabel)

        self.sentenceStructureLabel = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.sentenceStructureLabel.setFont(font)
        self.sentenceStructureLabel.setObjectName("sentenceStructureLabel")
        self.formLayout.setWidget(8, QtWidgets.QFormLayout.LabelRole, self.sentenceStructureLabel)

        self.impreciseFormLabel = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.impreciseFormLabel.setFont(font)
        self.impreciseFormLabel.setObjectName("impreciseFormLabel")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.LabelRole, self.impreciseFormLabel)

        self.image = QLabel(self.dockWidgetContents)
        self.image.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.image.setObjectName("image")
        self.formLayout.setWidget(10, QtWidgets.QFormLayout.FieldRole, self.image)

        self.color_legend = QLabel(self.dockWidgetContents)
        self.color_legend.setObjectName("color_legend")
        self.formLayout.setWidget(12, QtWidgets.QFormLayout.FieldRole, self.color_legend)

        self.checkBox = QtWidgets.QCheckBox(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.checkBox.setFont(font)
        self.formLayout.setWidget(12, QtWidgets.QFormLayout.LabelRole, self.checkBox)

        self.selectPDFButton = QtWidgets.QPushButton(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.selectPDFButton.setFont(font)
        self.selectPDFButton.setObjectName("selectPDFButton")
        self.formLayout.setWidget(13, QtWidgets.QFormLayout.LabelRole, self.selectPDFButton)

        self.label = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.FieldRole, self.label)
        self.label.setFixedWidth(150)  # set max length of path label
        self.formLayout.setWidget(13, QtWidgets.QFormLayout.FieldRole, self.label)

        self.generatePDFButton = QtWidgets.QPushButton(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        font.setBold(False)
        font.setWeight(50)
        self.generatePDFButton.setFont(font)
        self.generatePDFButton.setObjectName("generatePDFButton")
        self.formLayout.setWidget(14, QtWidgets.QFormLayout.LabelRole, self.generatePDFButton)

        self.overall_score_box = QtWidgets.QHBoxLayout()
        self.overall_score_box.setObjectName("overall_score_box")
        self.overall_score_label = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        self.overall_score_label.setFont(font)
        self.overall_score_label.setObjectName("overall_score_label")
        self.overall_score_box.addWidget(self.overall_score_label)
        self.overall_score_value = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(13)
        self.overall_score_value.setFont(font)
        self.overall_score_value.setObjectName("overall_score_value")
        self.overall_score_box.addWidget(self.overall_score_value)
        self.formLayout.setLayout(15, QtWidgets.QFormLayout.FieldRole, self.overall_score_box)

        self.overall_score_feedback = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Calibri")
        font.setPointSize(24)
        self.overall_score_feedback.setFont(font)
        self.overall_score_feedback.setObjectName("overall_score_feedback")
        self.formLayout.setWidget(17, QtWidgets.QFormLayout.FieldRole, self.overall_score_feedback)

        ui_e.overall_score_value.setVisible(False)
        ui_e.overall_score_label.setVisible(False)
        ui_e.overall_score_feedback.setVisible(False)

        self.dockWidget.setWidget(self.dockWidgetContents)
        self.dockWidget.setWindowTitle('Readability Settings')
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(1), self.dockWidget)

        self.actionShow_slider_window = QtWidgets.QAction(MainWindow)
        self.actionShow_slider_window.setObjectName("actionShow_slider_window")
        self.menuMenu.addAction(self.actionShow_slider_window)
        self.menubar.addAction(self.menuMenu.menuAction())

        self.overlay = Overlay(self.detailedView)
        self.overlay.move(400, 250)
        self.overlay.hide()
        # self.selectPDFButton.clicked.connect(self.overlay.show)

        # resize columns on startup
        self.resize_detail_view()

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)
        connect_buttons_thread = Thread(target=self.connect_buttons(), args=[])
        connect_buttons_thread.start()

        set_style_sheet(self.centralwidget)
        set_style_sheet(self.checkBox)
        set_style_sheet(self.dockWidget)
        set_style_sheet(self.detailedView)
        set_style_sheet(self.selectPDFButton)
        set_style_sheet(self.generatePDFButton)
        set_style_sheet(self.sentenceStructureSlider)
        set_style_sheet(self.sentenceLengthSlider)
        set_style_sheet(self.wordLengthSlider)
        set_style_sheet(self.vocabularySlider)
        set_style_sheet(self.nominalFormSlider)

    def connect_buttons(self):
        """
        Connect the pdf buttons with their processing methods.

        :return:    None
        """
        self.selectPDFButton.clicked.connect(self.set_pdf)
        if self.handler is None:
            self.generatePDFButton.setEnabled(False)
        self.generatePDFButton.clicked.connect(self.generate_pdf_with_button)

    def resizeEvent(self, event):
        self.overlay.resize(event.size())
        event.accept()

    def testPrint(self):
        """
        Prints a message when the generate-pdf-button is clicked.

        :return:    None
        """
        self.generatePDFButton.setEnabled(False)
        print("Generate PDF button clicked.")
        self.generatePDFButton.setEnabled(True)

    def retranslateUi(self, MainWindow):
        """
        Retranslate the view of the ui.

        :param MainWindow:  the main window of the ui
        :return:    None
        """
        _translate = QtCore.QCoreApplication.translate

        # identify cell selection and edit
        self.delegate = ItemDelegate(MainWindow)
        self.detailedView.setItemDelegate(self.delegate)
        self.detailedView.cellChanged.connect(self.cell_edited)
        self.detailedView.cellClicked.connect(self.cell_edit_start)
        self.actionShow_slider_window.triggered.connect(self.download_nltk)
        self.checkBox.stateChanged.connect(self.checkbox_clicked)

        MainWindow.setWindowTitle(_translate("MainWindow", "[Vis2] TextAnalyzer"))
        self.vocabularyLabel.setText(_translate("MainWindow", "Vocabulary Complexity (VC)"))
        self.wordLengthLabel.setText(_translate("MainWindow", "Word Length (WL)"))
        self.nominalFormLabel.setText(_translate("MainWindow", "Nominal Form (NF)"))
        self.sentenceLengthLabel.setText(_translate("MainWindow", "Sentence Length (SL)"))
        self.sentenceStructureLabel.setText(_translate("MainWindow", "Sentence Structure (SS)"))
        self.impreciseFormLabel.setText(_translate("MainWindow", "Imprecise Formulations (IF)"))
        self.selectPDFButton.setText(_translate("MainWindow", "       Select PDF file       "))
        self.generatePDFButton.setText(_translate("MainWindow", "       Generate PDF       "))
        self.checkBox.setText(_translate("MainWindow", "   Ignore references   "))
        self.label.setText(_translate("MainWindow", "No PDF file selected."))
        self.menuMenu.setTitle(_translate("MainWindow", "Menu"))
        self.actionShow_slider_window.setText(_translate("MainWindow", "Open NLTK Downloader"))
        item = self.detailedView.horizontalHeaderItem(0)
        item.setText(_translate("MainWindow", "Text sentence"))
        item = self.detailedView.horizontalHeaderItem(1)
        item.setText(_translate("MainWindow", "VC"))
        item = self.detailedView.horizontalHeaderItem(2)
        item.setText(_translate("MainWindow", "WL"))
        item = self.detailedView.horizontalHeaderItem(3)
        item.setText(_translate("MainWindow", "NF"))
        item = self.detailedView.horizontalHeaderItem(4)
        item.setText(_translate("MainWindow", "SL"))
        item = self.detailedView.horizontalHeaderItem(5)
        item.setText(_translate("MainWindow", "SS"))
        item = self.detailedView.horizontalHeaderItem(6)
        item.setText(_translate("MainWindow", "IF"))
        self.overall_score_label.setText(_translate("MainWindow", "Overall score"))
        self.overall_score_value.setText(_translate("MainWindow", "0.00"))
        self.overall_score_feedback.setText(_translate("MainWindow", "Neutral"))
        pixmap = QPixmap('../res/scala.png')
        pixmap_table = QPixmap('../res/color.png')
        pixmap_scala = QPixmap('../res/color_small.png')
        self.color_legend.setPixmap(
            self.paint_text(pixmap_scala, QPoint(4, 12), QPoint(124, 12), 'easy', 'hard', QColor(255, 255, 255)))
        self.img_table.setPixmap(self.paint_text(pixmap, QPoint(4, 32), QPoint(162, 32), '0', '100', QColor(0, 0, 0)))
        self.image.setPixmap(pixmap)
        self.image.setScaledContents(True)
        self.img_table.setPixmap(pixmap_table)
        self.img_table.setScaledContents(True)

    @staticmethod
    def paint_text(pixmap, start_pos, end_pos, start_text, end_text, color):
        """
        Paint two text-strings on a pixmap.

        :param pixmap:  the pixmap to paint the text on
        :param start_pos:   the position of the first text-string
        :param end_pos:     the position of the second text-string
        :param start_text:  the first text-string to paint
        :param end_text:    the second text-string to paint
        :param color:   the color used for painting
        :return:    the pixmap with the text painted on
        """
        painter = QPainter()
        painter.begin(pixmap)
        painter.setPen(color)
        painter.drawText(start_pos, start_text)
        painter.drawText(end_pos, end_text)
        painter.end()
        return pixmap

    @staticmethod
    def download_nltk():
        """
        Open the NLTK downloader.

        :return:    None
        """
        nltk.download()

    def checkbox_clicked(self):
        """
        Called when the reference-checkbox is clicked to update the view.

        :return:    None
        """
        if len(pdf_path) > 0:
            self.process_file(pdf_path)

    def cell_edited(self):
        """
        Called when a cell in the detail view is changed. It updates all scores and colors of the sentence edited.

        :return:    None.
        """
        # recognize change only when cell is edited afterwards
        global initialized
        if initialized:
            print("Edited row %d and column %d." % (cell_row, cell_column))
            item = self.detailedView.item(cell_row, cell_column)
            print("New sentence text: ", item.text())

            initialized = 0
            global sent
            # update sentence
            sentence = []

            # always set the new text in the sentence column if it is written somewhere else
            if cell_column > 0:
                self.detailedView.setItem(cell_row, 0, QTableWidgetItem(str(item.text())))

            for x in range(len(sent)):
                if sent[x].sentence == cell_text:
                    sentence = sent[x]
                    sent[x].sentence = item.text()
            new_score = analyze_single_sentence(item.text())
            print("New sentence scores: ", new_score)

            # update entry in score matrix
            global scores
            for c in range(len(scores[0])):
                scores[cell_row][c] = new_score[0][c]
            print(scores)

            # update score
            for j in range(len(new_score[0])):
                self.detailedView.setItem(cell_row, j + 1,
                                          QTableWidgetItem(str("%.2f" % round(new_score[0][j], 2))))

                interpolate_score_color(new_score[0][j], sentence)
                update_score_color(cell_row, j + 1, sentence)

            sentence_score = calculate_sentence_score(new_score, 0)
            interpolate_sentence_color(sentence_score, sentence)
            update_sentence_color(cell_row, sentence)

            self.detailedView.resizeRowsToContents()
            initialized = 1

            # calculate new overall score
            sum_score = 0
            for j in range(len(scores)):
                sent_score = calculate_sentence_score(scores, j)
                sum_score = sum_score + sent_score
            update_overall_score(sum_score, len(scores))

            # after editing text, disable pdf button until an update
            self.generatePDFButton.setEnabled(False)

    def cell_edit_start(self, row, column):
        """
        Called when the user selects a cell to save the cell data.

        :param row: the row of the cell selected
        :param column:  the column of the cell selected
        :return:    None
        """
        global cell_row, cell_column, cell_text

        print("Clicked on row %d and column %d." % (row, column))
        item = self.detailedView.item(row, 0)
        # print(item.text())
        cell_row = row
        cell_column = column
        cell_text = item.text()

    def set_pdf(self):
        """
        Called when the user clicks the select-pdf button. It opens a pdf-file selection dialog.
        Further, it initiates the file processing.

        :return:    None
        """
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(None, "Select PDF", "", "PDF Files (*.pdf *.PDF)")
        if not os.path.exists(file_name):
            return

        if file_name:  # if file is selected
            global pdf_path
            pdf_path = file_name

            print("PDF selected: " + file_name)
            self.label.setText(ntpath.basename(file_name))

            pdf_handler_thread = Thread(target=self.process_file(file_name), args=[])
            pdf_handler_thread.start()

    def process_file(self, file_name):
        """
        Initiate the sentence extraction and set the scores for all sentence data.

        :param file_name:   the name of the file selected
        :return:    None
        """
        waiting_indicator_thread = Thread(target=self.overlay.show(), args=[])
        waiting_indicator_thread.start()

        handler = PdfHandler(file_name, self.checkBox.isChecked())
        sentences = handler.getSentence()
        global initialized, sent
        initialized = 0
        sent = sentences

        self.detailedView.setRowCount(len(sentences))
        for i, s in enumerate(sentences):
            self.detailedView.setItem(i, 0, QTableWidgetItem(s.sentence))

        global scores
        scores = text_analysis(sentences)
        print(scores)

        # set scores
        sum_score = 0
        for k, s in enumerate(sentences):
            # color of sentence
            self.detailedView.item(k, 0).setBackground(
                QtGui.QColor(s.color[0] * 255, s.color[1] * 255, s.color[2] * 255))

            for j in range(len(scores[k])):
                self.detailedView.setItem(k, j + 1, QTableWidgetItem(str("%.2f" % round(scores[k][j], 2))))

                interpolate_score_color(scores[k][j], s)
                update_score_color(k, j + 1, s)

            sentence_score = calculate_sentence_score(scores, k)
            interpolate_sentence_color(sentence_score, s)
            update_sentence_color(k, s)
            sum_score = sum_score + sentence_score

        self.detailedView.resizeRowsToContents()
        update_overall_score(sum_score, len(sent))
        initialized = 1

        # update handler
        self.generatePDFButton.setEnabled(True)
        self.handler = handler

    def generate_pdf_with_button(self):
        """
        Generate a pdf, highlight it, and open it with the standard program of the user.

        :return:    None
        """
        self.generatePDFButton.setEnabled(False)
        highlight_pdf_thread = Thread(target=self.handler.generateHighlightedPdf(), args=[])
        highlight_pdf_thread.start()
        highlight_pdf_thread.join()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        my_list = os.listdir(dir_path + '\\..\\..\\pdf')
        webbrowser.open(dir_path + '\\..\\..\\pdf\\' + my_list[-1])

    def resize_detail_view(self):
        """
        Resize the columns of the detail view.

        :return:    None
        """
        header = self.detailedView.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        lis = 60
        self.detailedView.setColumnWidth(1, lis)
        self.detailedView.setColumnWidth(2, lis)
        self.detailedView.setColumnWidth(3, lis)
        self.detailedView.setColumnWidth(4, lis)
        self.detailedView.setColumnWidth(5, lis)
        self.detailedView.setColumnWidth(6, lis)


class ItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    Re-implement the edit event to detect changes in the detail view.

    """
    cellEditingStarted = QtCore.pyqtSignal(int, int)

    def createEditor(self, parent, option, index):
        """
        Re-implement function to customize the edit behavior.

        """
        result = super(ItemDelegate, self).createEditor(parent, option, index)
        if result:
            self.cellEditingStarted.emit(index.row(), index.column())
        return result


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())
