import os
import shutil

import fitz
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LAParams, LTTextBox, LTTextLine, LTChar
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from datetime import datetime


class CharData:
    """
    Hold properties of characters, which are extracted from the pdf data.
    """

    def __init__(self, character, bb0, bb1, bb2, bb3):
        """
        This is the constructor.

        :param character:
        :param bb0: The distance from the left of the page to the left edge of the box.
        :param bb1: The distance from the bottom of the page to the lower edge of the box.
        :param bb2: The distance from the left of the page to the right edge of the box.
        :param bb3: The distance from the bottom of the page to the upper edge of the box.
        """
        self.character = character
        self.bb0 = bb0  # the distance from the left of the page to the left edge of the box.
        self.bb1 = bb1  # the distance from the bottom of the page to the lower edge of the box.
        self.bb2 = bb2  # the distance from the left of the page to the right edge of the box.
        self.bb3 = bb3  # the distance from the bottom of the page to the upper edge of the box.


class SentenceData:
    def __init__(self, sentence, rectList, pageNum):  # , lrx, lry):
        """
        Hold properties of generated sentences, which are generated from extracted characters of pdf.

        :param sentence:    The string data for each sentence.
        :param rectList:    The bounding boxes of the sentences in the pdf.
        :param pageNum: The page of pdf, which the sentence belongs to.
        """

        self.sentence = sentence
        self.color = (0, 1, 1)
        self.rectList = rectList
        self.annotText = ""
        self.pageNum = pageNum

    def setColor(self, color):
        """
        Set the annotation color for the sentence.

        :param color:   The color
        :return:    None
        """
        self.color = color

    def setRectList(self, offset):
        """
        Define original coordinate for pdf.

        :param offset:  The offsets from original coordinate.
        :return:    None
        """
        for rect in self.rectList:
            rect.y0 = offset - rect.y0
            rect.y1 = offset - rect.y1

    def setAnnotation(self, annotText):
        """
        Set annotate text for pdf. The text will be annotated in the pdf data.

        :param annotText:   Annotate text.
        :return:    None
        """
        self.annotText = annotText


class PageData:
    """
    Hold the data of each page including sentences and characters.
    """

    def __init__(self):
        """
        Constructor.
        """
        self.charDataList = []
        self.sentenceList = []

    def addChar(self, c):
        """
        Add character data to the list.

        :param c:   Added character.
        :return:    None
        """
        self.charDataList.append(c)

    def addSentence(self, s):
        """
        Add sentence data to the list.

        :param s:   Added sentence.
        :return:    None
        """
        self.sentenceList.append(s)


class PdfHandler:
    """
    Handle whole pdf data.
    """

    def __init__(self, pdfPath, refIgn):
        """
        Constructor.

        :param pdfPath: Path to the pdf from which system extract characters.
        """
        print("PdfHandler called.")
        self.refIgn = refIgn
        self.pdfPath = pdfPath
        # self.pdfPath = os.getcwd() + '/../../pdf/Oelke_visual-readability.pdf' # Delete Me
        self.PageDataList = []
        # print('set the flag 0!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        self.flag = 0
        self.isGenerated = 0
        self.textExtracWithCoord()
        self.makeSentence()

    def textExtracWithCoord(self):
        """
        Extract each character from pdf data. The character and its coordinates are extracted.

        :return:    None
        """
        fp = open(self.pdfPath, 'rb')
        manager = PDFResourceManager()
        laparams = LAParams()
        # laparams.detect_vertical = True

        device = PDFPageAggregator(manager, laparams=laparams)
        interpreter = PDFPageInterpreter(manager, device)
        pages = PDFPage.get_pages(fp)

        for page in pages:
            page_data = PageData()
            print('Processing next page... ')
            interpreter.process_page(page)
            layout = device.get_result()
            for lobj in layout:
                if isinstance(lobj, LTTextBox):
                    # x, y, text = lobj.bbox[0], lobj.bbox[3], lobj.get_text()
                    # print('At %r is text: %s' % ((x, y), text))
                    # print('LTText!!!!!')
                    for obj in lobj._objs:
                        if isinstance(obj, LTTextLine):
                            # x, y, text = lobj.bbox[0], lobj.bbox[3], obj.get_text()
                            # print(repr('At %r is text: %s' % ((x, y), text)))
                            for char in obj._objs:
                                if isinstance(char, LTChar):
                                    x0, y0, x1, y1, ch = char.bbox[0], char.bbox[1], char.bbox[2], char.bbox[
                                        3], char.get_text()
                                else:
                                    x0, y0, x1, y1, ch = 0, 0, 0, 0, char.get_text()
                                page_data.addChar(CharData(ch, x0, y0, x1, y1))
                        # activate below if you want to have '\n' information
                        # self.charDataList.append(CharData('\n', 0, 0, 0, 0))
            self.PageDataList.append(page_data)
        fp.close()
        device.close()

    def makeSentence(self):
        """
        Make sentences from extracted characters.

        :return:    None
        """
        # make list of sentences
        sent = ''
        rectList = []
        current_rect = fitz.Rect()
        endFlag = 0

        def end_sentence(rectList, current_rect, sent, pageNum):
            """
            An intrinsic function, which initializes sentence parameters.

            :param rectList:    The bounding boxes for the sentence.
            :param current_rect:    The bounding box of currently working line.
            :param sent:    The sentence.
            :param pageNum: The number of the page.
            :return:    None
            """
            if current_rect.getArea() != 0:
                rectList.append(current_rect)
            page.addSentence(SentenceData(sent, rectList, pageNum))
            rectList = []
            current_rect = fitz.Rect()
            sent = ''
            return rectList, current_rect, sent

        for i, page in enumerate(self.PageDataList):
            if endFlag == 1:
                break
            print("+++++++++++++++++++++++++++++++++++++++++")
            for j, a in enumerate(page.charDataList):
                # this if sentence is to get text coordinates with rectangle
                if a.character != '\n':
                    sent = sent + a.character
                    if current_rect.getArea() == 0:
                        current_rect = fitz.Rect(a.bb0, a.bb1, a.bb2, a.bb3)
                    elif a.bb2 > current_rect.x1:
                        current_rect = fitz.Rect(current_rect.x0, current_rect.y0, a.bb2, current_rect.y1)
                else:
                    if j - 1 >= 0:
                        if page.charDataList[j - 1].character == '-':
                            sent = sent[:-1]
                        elif page.charDataList[j - 1].character == ' ':
                            pass
                        else:
                            sent = sent + ' '
                    try:
                        before = page.charDataList[j - 1]
                        after = page.charDataList[j + 1]
                        if abs(after.bb3 - before.bb1) > 0.1 * abs(before.bb1 - before.bb3):
                            # if (abs(current_rect.y0-current_rect.y1)<abs(before.bb1-before.bb3))*0.8 or (abs(before.bb1 - before.bb3)< abs(current_rect.y0-current_rect.y1))*0.8:
                            rectList, current_rect, sent = end_sentence(rectList, current_rect, sent, i)
                    except:
                        print("This was the first letter.")
                    rectList.append(current_rect)
                    current_rect = fitz.Rect()

                # In this if sentence, definition of a sentence is defined
                if j == len(page.charDataList) - 1:
                    # 1. The last sentence of a page counted as a sentence even if it does not end with '.'
                    print('=================')
                    rectList, current_rect, sent = end_sentence(rectList, current_rect, sent, i)
                elif a.character == '.':
                    # 2. A sentence ends with '.'
                    if page.charDataList[j - 2].character == 'a' and page.charDataList[j - 1].character == 'l':
                        # 2.1 except 'et al.'
                        continue
                    elif page.charDataList[j - 1].character.isdigit():
                        # 2.2 if any number comes before '.' except
                        continue
                    elif (page.charDataList[j - 1].character == 'e' and page.charDataList[j + 1].character == 'g') \
                            or (
                            page.charDataList[j - 2].character == '.' and page.charDataList[j - 1].character == 'g'):
                        # 2.3 except 'e.g.'
                        continue
                    elif (page.charDataList[j - 1].character == 'i' and page.charDataList[j + 1].character == 'e') \
                            or (
                            page.charDataList[j - 2].character == '.' and page.charDataList[j - 1].character == 'e'):
                        # 2.4 except 'i.e.'
                        continue
                    elif page.charDataList[j - 2].character == ' ':
                        # 2.5 except Name
                        continue
                    elif page.charDataList[j + 1].character != ' ' and page.charDataList[j + 1].character != '\n':
                        # 2.6 except '.' which has not blank after that
                        continue
                    else:
                        rectList, current_rect, sent = end_sentence(rectList, current_rect, sent, i)
                elif a.character == ':':
                    try:
                        if page.charDataList[j + 1].character == '\n':
                            rectList, current_rect, sent = end_sentence(rectList, current_rect, sent, i)
                    except:
                        print("This was the first letter.")
                        continue
                elif a.character == 'R' and self.refIgn:
                    try:
                        if (page.charDataList[j + 1].character == 'e' or page.charDataList[j + 1].character == 'E') \
                                and (
                                page.charDataList[j + 2].character == 'f' or page.charDataList[j + 2].character == 'F') \
                                and (
                                page.charDataList[j + 3].character == 'e' or page.charDataList[j + 3].character == 'E') \
                                and (
                                page.charDataList[j + 4].character == 'r' or page.charDataList[j + 4].character == 'R') \
                                and (
                                page.charDataList[j + 5].character == 'e' or page.charDataList[j + 5].character == 'E') \
                                and (
                                page.charDataList[j + 6].character == 'n' or page.charDataList[j + 6].character == 'N') \
                                and (
                                page.charDataList[j + 7].character == 'c' or page.charDataList[j + 7].character == 'C') \
                                and (
                                page.charDataList[j + 8].character == 'e' or page.charDataList[j + 8].character == 'E') \
                                and page.charDataList[j + 9].character == '\n':
                            endFlag = 1
                            break
                        elif (page.charDataList[j + 1].character == 'e' or page.charDataList[j + 1].character == 'E') \
                                and (
                                page.charDataList[j + 2].character == 'f' or page.charDataList[j + 2].character == 'F') \
                                and (
                                page.charDataList[j + 3].character == 'e' or page.charDataList[j + 3].character == 'E') \
                                and (
                                page.charDataList[j + 4].character == 'r' or page.charDataList[j + 4].character == 'R') \
                                and (
                                page.charDataList[j + 5].character == 'e' or page.charDataList[j + 5].character == 'E') \
                                and (
                                page.charDataList[j + 6].character == 'n' or page.charDataList[j + 6].character == 'N') \
                                and (
                                page.charDataList[j + 7].character == 'c' or page.charDataList[j + 7].character == 'C') \
                                and (
                                page.charDataList[j + 8].character == 'e' or page.charDataList[j + 8].character == 'E') \
                                and (
                                page.charDataList[j + 9].character == 's' or page.charDataList[j + 9].character == 'S') \
                                and page.charDataList[j + 10].character == '\n':
                            endFlag = 1
                            break
                    except:
                        print("This was not last")

    def getSentence(self):
        """
        Gets all sentences of the pdf data.

        :return:    Whole sentences.
        """
        sentences = []
        for page in self.PageDataList:
            for s in page.sentenceList:
                sentences.append(s)
        return sentences

    def generateHighlightedPdf(self, removeRef=False):
        """
        Generate highlighted pdf with respect to each color and annotating text of it.

        :return:    None
        """
        pdfPathSave = os.getcwd() + '/../../pdf/result_%s.pdf' % (datetime.now().strftime("%Y%m%d%H%M%S"))
        shutil.copy(self.pdfPath, pdfPathSave)
        doc = fitz.open(pdfPathSave)
        offset = 0
        endFlag = 0
        for i, page in enumerate(doc):
            for t in self.PageDataList[i].sentenceList:
                if (t.sentence[0:10] == 'References' or t.sentence[0:10] == 'REFERENCES' or t.sentence[0:9] == 'Reference' or  t.sentence[0:9] == 'REFERENCE') and False:
                    endFlag = 1

                if self.flag == 0:
                    t1 = page.searchFor(t.sentence[0:6])
                    t2 = t.rectList
                    # print(len(t1))
                    # print(len(t2))
                    offset = (t2[0].y0 if (len(t2) > 0) else 0) + (t1[0].y1 if (len(t1) > 0) else 0)
                    self.flag = 1
                if not self.isGenerated:
                    t.setRectList(offset)
                text_instances = t.rectList  # page.searchFor(t.sentence)
                # a = self.PageDataList[1].charDataList[2]
                # b = self.PageDataList[1].charDataList[128]
                # p=845
                # text_instances = [fitz.Rect(b.bb0, p-b.bb1, b.bb2, p-b.bb3),fitz.Rect(a.bb0, p-a.bb1, a.bb2, p-a.bb3)]
                # print(text_instances)
                if (endFlag == 0):
                    for inst in text_instances:
                        highlight = page.addHighlightAnnot(inst)
                        highlight.setColors({"stroke": t.color})
                        highlight.update()
                    if t.annotText != "":
                        page.addTextAnnot((text_instances[0].x1, text_instances[0].y1), t.annotText)
        self.isGenerated = 1
        # self.doc.save(self.pdfPathSave, garbage=4, deflate=True, clean=True)
        # print(doc)
        doc.save(pdfPathSave, incremental=True)


if __name__ == '__main__':
    pdfPath = "aaa"  # path to the pdf
    pdfHandler = PdfHandler(pdfPath)
    pdfHandler.test()
    # contents = pdfHandler.textExtract()
    # print(contents)
    # print(len(pdfHandler.charDataList))
    sts = pdfHandler.getSentence()
    for s in sts:
        print(s.sentence)
